from abc import ABCMeta, abstractmethod
from collections import deque

from .variant import Variant
from .model import Model, Memory
from .isa import InstructionNOP
from .program import Program
from .types import RVFISignals, TraceIntegerRegister, Register
from .code import decode


class GoldenException(Exception):
    pass


class GoldenProgramEndException(Exception):
    pass


class Golden(metaclass=ABCMeta):
    def __init__(self, variant: Variant):
        self.model = Model(variant)
        self.memory = Memory()
        self.program = None

    @abstractmethod
    def fetch(self, pc):
        pass


class GoldenUnbuffered(Golden):
    def __init__(self, variant: Variant, *, pc: int = 0):
        super().__init__(variant)
        self.reset(pc=pc)

    def fetch(self, pc):
        if self.pc != pc:
            raise GoldenException("Unexpected fetch PC: {}, expected {}".format(pc, self.pc))
        try:
            insn = self.program[pc >> 2]
        except IndexError:
            raise GoldenProgramEndException()
        self.issued.append(insn)
        self.pc += 4
        return insn

    def commit(self, trace, *, insn = None):
        if not isinstance(trace, list):
            trace = [trace]

        # Test if any other exceptions are issued and get in order
        try:
            exp = self.issued.popleft()
        except IndexError:
            # If not we may be able to verify this is a NOP, then its okay. If no NOP, its probably okay too..
            if insn is not None:
                if insn != InstructionNOP():
                    raise GoldenProgramEndException()
            raise GoldenProgramEndException()

        # If we got an instruction check if it matches
        if insn is not None:
            if exp != insn:
                raise GoldenException("Expected instruction: {}, expected {}".format(insn, exp))

        # Execute the expected instruction and verify the state is the same
        exp_trace = self.model.issue(exp)
        if not self.model.check(trace):
            raise GoldenException("Unexpected state change: {}, expected: {}".format(",".join([str(t) for t in trace]),
                                                                                     ",".join([str(t) for t in exp_trace])))

    def reset(self, *, pc: int = 0):
        self.model.reset(pc=pc)
        self.pc = pc
        self.issued = deque()

    def load_program(self, pgm: Program):
        self.program = pgm.insns


def traces_from_rvfi(rvfi: RVFISignals) -> list:
    insn = decode(rvfi.insn)
    if rvfi.valid != 1:
        return []
    t = []
    if rvfi.rd_addr == 0 and rvfi.rd_wdata != 0:
        raise ValueError("rd[0] cannot be written by core")
    if rvfi.rd_addr != 0:
        if "rd" in insn.__dict__:
            reg = Register(32)
            reg.set(rvfi.rd_wdata)
            t.append(TraceIntegerRegister(rvfi.rd_addr, reg))

    return t


from .model import Model, Memory, TerminateException

import struct

class Simulator:
  def __init__(self, model):
    self.model = model
    self.program = []

  def load_program(self, program, *, address=0):
    self.program = [i for i in program]

  def load_data(self, data = "", *, address=0):
    mem = self.model.state.memory.memory
    for a in range(int(len(data)/4)):
      mem[a] = struct.unpack("<L", data[a*4:(a+1)*4])[0]

  def run(self, *, pc=0):
    self.model.reset(pc=pc)
    cnt = 0
    while True:
      try:
        self.model.issue(self.program[int(self.model.state.pc)>>2])
        cnt += 1
      except TerminateException as exc:
        assert exc.returncode == 0
        return cnt
      except IndexError as exc:
        return cnt
    return cnt

  def dump_data(self, *, address=0, size=None):
    data = b""
    mem = self.model.state.memory.memory
    for a in mem:
      if a >= address and (size is None or a < address + size):
        data += struct.pack("<L", mem[a])
    return data



# Copyright Stefan Wallentowitz
# Licensed under the MIT License, see LICENSE for details.
# SPDX-License-Identifier: MIT
'''
TODO: Description
'''

import argparse
import re
from collections import namedtuple

Extension = namedtuple("Extension", ["name", "description", "implies"])


class Variant:
    '''
    RISC-V ISA Variants
    '''
    stdext = {
        "M": Extension("M", "Integer Multiplication and Division", []),
        "A": Extension("A", "Atomics", []),
        "F": Extension("F", "Single-Precision Floating-Point", ["Zicsr"]),
        "D": Extension("D", "Double-Precision Floating-Point", ["F"]),
        "C": Extension("C", "16-bit Compressed Instructions", [])
    }
    stdextZ = {
        "Zicsr": Extension("Zicsr", "Control and Status Register Access", []),
        "Zifencei": Extension("Zifencei", "Instruction-Fetch Fence", [])
    }
    regex = re.compile(
        r"^RV(32|64|128)(A|I|E|G)({})*((?:(?:Z|X).+)(?:_(?:(?:Z|X).*))*)?$".
        format("|".join(list(stdext.keys()) + ["G"])))
    G_expand = ["I", "M", "A", "F", "D", "Zicsr", "Zifencei"]

    def __init__(self, name: str, *, custext=None):
        self.custext = {}
        for ext in custext or []:
            self.custext[ext.name] = ext
        self.name = name
        match = self.regex.match(name.upper())
        assert match, "Invalid variant string '{}'".format(name)
        self.extensions = set()
        self.xlen = int(match.group(1))
        self.baseint = match.group(2)
        if self.baseint == "G":
            self.extensions |= set(Variant.G_expand)
            self.baseint = "I"
        assert (self.baseint == "I"
                or self.xlen == 32), "E base integer is only valid for 32-bit"
        if match.group(3):
            for ext in match.group(3):
                if ext == "G":
                    assert (self.baseint == "I"), "G is not defined for I"
                    self.extensions |= set(Variant.G_expand)
                else:
                    self.extensions |= set([ext] + Variant.stdext[ext].implies)
        if match.group(4):
            for ext in match.group(4).split("_"):
                if ext[0] == "Z":
                    self.extensions |= set(
                        [ext[0]+ext[1:].lower()] + Variant.stdextZ["Z" + ext[1:].lower()].implies)
                elif ext[0] == "X":
                    self.extensions |= set(
                        [ext[0]+ext[1:].lower()] + self.custext["X" + ext[1:].lower()].implies)

    def __str__(self):
        return self.name

    def describe(self):
        '''
        Describe what this RISC-V ISA Variant defines
        '''
        desc = self.name + "\n"
        desc += "  XLEN={}, {} integer registers ({})\n".format(
            self.xlen, "16" if self.baseint == "E" else "32", self.baseint)
        desc += "  Extensions:\n"
        if len(self.extensions) == 0:
            desc += "    None\n"
        else:
            for entry in sorted(self.extensions):
                ext = None
                if entry in Variant.stdext:
                    ext = Variant.stdext[entry]
                elif "Z" + entry[1:].lower() in Variant.stdextZ:
                    ext = Variant.stdextZ["Z" + entry[1:].lower()]
                elif "X" + entry[1:].lower() in self.custext:
                    ext = self.custext["X" + entry[1:].lower()]
                assert ext, "Cannot find extension {}".format(entry)
                desc += "    {:8} {}\n".format(ext.name, ext.description)
        return desc

    def __eq__(self, other):
        '''
        Check if two Variants are equal
        '''
        return (self.xlen == other.xlen and self.baseint == other.baseint
                and self.extensions == other.extensions)

    def __le__(self, other):
        '''
        Check if Variant is a subset of another
        '''
        return (self.xlen == other.xlen and self.baseint == other.baseint
                and self.extensions <= other.extensions)

    def __add__(self, other):
        '''
        Add extensions to a Variant
        '''
        if isinstance(other, str):
            self.extensions.add(other)
        else:
            if isinstance(other, Extension):
                other = [other]
            assert isinstance(other,
                              list), "Requires Extension or list of Extension"
            for ext in other:
                self.custext[ext.name] = ext
                self.extensions.add(ext.name)
                for i in ext.implies:
                    self.extensions.add(i)
        return self


# Convenience
RV32I = Variant("RV32I")
RV32E = Variant("RV32E")
RV32IZicsr = Variant("RV32IZicsr")
RV32IZifencei = Variant("RV32IZifencei")
RV32IM = Variant("RV32IM")
RV32IC = Variant("RV32IC")
RV64I = Variant("RV64I")
RV64G = Variant("RV64G")
RV64GC = Variant("RV64GC")
RV128I = Variant("RV128I")
RV32A = Variant("RV32A")


def describe_argparser():
    """
    Return an argument parser for the describe function
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("variant", help="Variant to describe, example: RV64GC")
    return parser


def describe():
    '''
    Function called by command line tool

    Describe a RISC-V ISA Variant based on the name
    '''
    parser = describe_argparser()
    args = parser.parse_args()
    variant = Variant(args.variant)

    print(variant.describe())



from random import randrange
import sys

from .variant import *
from .types import Register, RegisterFile, TracePC, TraceIntegerRegister, TraceMemory
from .program import Program

class TerminateException(Exception):
    """
    Exception that signal the termination of the program
    """
    def __init__(self, returncode):
        super().__init__()
        self.returncode = returncode

# todo: raise exception on memory misalignment
class State(object):
    single_regs = []

    def __init__(self, variant: Variant):
        self.variant = variant
        intregs = 32 if variant.baseint == "I" else 16
        self.intreg = RegisterFile(intregs, 32, {0: 0x0})
        self.pc = Register(32)
        self.pc_update = Register(32)
        self.memory = Memory()
        self.initialized = True
        # Used for tracking atomic accesses
        self.reservations = []

    def randomize(self):
        self.intreg.randomize()

    def changes(self):
        c = self.intreg.changes()
        if self.pc_update.value != self.pc.value + 4:
            c.append(TracePC(self.pc_update.value))
        c += self.memory.changes()
        return c

    def commit(self):
        self.intreg.commit()
        self.pc.set(self.pc_update.value)
        self.memory.commit()

    def reset(self, pc = 0):
        self.pc.set(pc)

    def __setattr__(self, key, value):
        if key in self.single_regs and key in self.__dict__:
            getattr(self, key).update(value)
        elif key == "pc" and "pc_update" in self.__dict__:
            self.pc_update.set(value)
        else:
            super().__setattr__(key, value)

    def __str__(self):
        return "{}".format(self.intreg)

    def atomic_acquire(self, address):
        if not address in self.reservations:
            self.reservations.append(address)

    def atomic_release(self, address):
        if address in self.reservations:
            self.reservations.remove(address)

    def atomic_reserved(self, address):
        return address in self.reservations

class Memory(object):
    def __init__(self, *, base: int = 0, size: int = 2^32):
        self.base = base
        self.size = size
        self.memory = {}
        self.memory_updates = []

    def lb(self, address):
        word = address >> 2
        offset = address % 4
        if word not in self.memory:
            self.memory[word] = randrange(0, 1 << 32)
        return (self.memory[word] >> (offset*8)) & 0xff

    def lh(self, address):
        word = address >> 2
        offset = (address >> 1) % 2
        if word not in self.memory:
            self.memory[word] = randrange(0, 1 << 32)
        return (self.memory[word] >> (offset*16)) & 0xffff

    def lw(self, address):
        word = address >> 2
        if word not in self.memory:
            self.memory[word] = randrange(0, 1 << 32)
        return self.memory[word]

    def sb(self, address, data):
        self.memory_updates.append(TraceMemory(TraceMemory.GRANULARITY.BYTE, address, int(data) & 0xFF))

    def sh(self, address, data):
        self.memory_updates.append(TraceMemory(TraceMemory.GRANULARITY.HALFWORD, address, int(data) & 0xFFFF))

    def sw(self, address, data):
        self.memory_updates.append(TraceMemory(TraceMemory.GRANULARITY.WORD, address, int(data) & 0xFFFFFFFF))

    def changes(self):
        return self.memory_updates

    def commit(self):
        for update in self.memory_updates:
            address = update.addr
            base = address >> 2
            offset = address & 0x3
            if base not in self.memory:
                self.memory[base] = randrange(0, 1 << 32)
            data = update.data
            if update.gran == TraceMemory.GRANULARITY.BYTE:
                mask = ~(0xFF << (offset*8)) & 0xFFFFFFFF
                data = (self.memory[base] & mask) | (data << (offset*8))
            self.memory[base] = data

        self.memory_updates = []

class Environment:
    def call(self, state: State):
        pass

class Model(object):
    def __init__(self, variant: Variant, *, environment: Environment = None, verbose = False, asm_width=20):
        self.state = State(variant)
        self.environment = environment if environment is not None else Environment()
        self.verbose = verbose
        if self.verbose is not False:
            self.verbose_file = sys.stdout if verbose is True else open(self.verbose, "w")
        self.asm_tpl = "{{:{}}} | [{{}}]\n".format(asm_width)

    def issue(self, insn):
        self.state.pc += 4
        expected_pc = self.state.pc
        insn.execute(self)

        trace = self.state.changes()
        if self.verbose is not False:
            self.verbose_file.write(self.asm_tpl.format(str(insn), ", ".join([str(t) for t in trace])))
        self.state.commit()
        return trace

    def execute(self, insn):
        if isinstance(insn, Program):
            insn = insn.insns
        elif not isinstance(insn, list):
            insn = [insn]

        try:
            for i in insn:
                self.issue(i)
        except TerminateException as ex:
            assert ex.returncode == 0

    def randomize(self):
        self.state.randomize()

    def reset(self, *, pc: int = 0):
        self.state.reset(pc)

    def check(self, traces, exp=None):
        for t in traces:
            if isinstance(t, TraceIntegerRegister):
                if int(self.state.intreg[t.id]) != int(t.value):
                    return False
        return True



from riscvmodel.program.tests import *
from riscvmodel.program.tests_atomic import *
from riscvmodel.model import Model
from riscvmodel.variant import RV32I, RV32A

import pytest


def check_model(model: Model, check: dict):
    diff = []
    for key, value in check.items():
        if isinstance(key, int):
            if model.state.intreg[key].unsigned() != value:
                diff.append("x{} mismatch: expected {:08x}, got {:08x}".format(key, value, model.state.intreg[key].unsigned()))
    if len(diff) > 0:
        msg = "Check failed\n{}".format("\n  ".join(diff))
        pytest.fail(msg)


def test_model_addi():
    pgm = ADDITest()
    m = Model(RV32I)
    m.execute(pgm)
    check_model(m, pgm.expects())


def test_model_lr_sc():
    """Test if load-reserved and store-conditional sequence works"""
    pgm = LRSCTest()
    m   = Model(RV32A)
    m.execute(pgm)
    check_model(m, pgm.expects())


# Copyright Stefan Wallentowitz
# Licensed under the MIT License, see LICENSE for details.
# SPDX-License-Identifier: MIT
"""
Instructions
"""

from .isa import *
from .variant import *
from .model import State

@isa("lui", RV32I, opcode=0b0110111)
class InstructionLUI(InstructionUType):
    """
    The Load Upper Immediate (LUI) instruction loads the given immediate (unsigned 20 bit) to the upper 20 bit
    of the destination register. The lower bits are set to zero in the destination register. This instruction
    can be used to efficiently form constants, as a sequence of LUI and ORI for example.
    """
    def execute(self, model: Model):
        model.state.intreg[self.rd] = (self.imm << 12)


@isa("auipc", RV32I, opcode=0b0010111)
class InstructionAUIPC(InstructionUType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.pc + (self.imm << 12)


@isa("jal", RV32I, opcode=0b1101111)
class InstructionJAL(InstructionJType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.pc + 4
        model.state.pc += self.imm


@isa("jalr", RV32I, opcode=0b1100111, funct3=0b000)
class InstructionJALR(InstructionIType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.pc + 4
        model.state.pc = model.state.intreg[self.rs1] + self.imm


@isa("beq", RV32I, opcode=0b1100011, funct3=0b000)
class InstructionBEQ(InstructionBType):
    def execute(self, model: Model):
        # todo: problem with __cmp__
        if model.state.intreg[self.rs1].value == model.state.intreg[self.rs2].value:
            model.state.pc = model.state.pc + self.imm


@isa("bne", RV32I, opcode=0b1100011, funct3=0b001)
class InstructionBNE(InstructionBType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1].value != model.state.intreg[self.rs2].value:
            model.state.pc = model.state.pc + self.imm


@isa("blt", RV32I, opcode=0b1100011, funct3=0b100)
class InstructionBLT(InstructionBType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1].value < model.state.intreg[self.rs2].value:
            model.state.pc = model.state.pc + self.imm


@isa("bge", RV32I, opcode=0b1100011, funct3=0b101)
class InstructionBGE(InstructionBType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1].value >= model.state.intreg[self.rs2].value:
            model.state.pc = model.state.pc + self.imm


@isa("bltu", RV32I, opcode=0b1100011, funct3=0b110)
class InstructionBLTU(InstructionBType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1].unsigned() < model.state.intreg[
                self.rs2].unsigned():
            model.state.pc = model.state.pc + self.imm


@isa("bgeu", RV32I, opcode=0b1100011, funct3=0b111)
class InstructionBGEU(InstructionBType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1].unsigned() >= model.state.intreg[
                self.rs2].unsigned():
            model.state.pc = model.state.pc + self.imm


@isa("lb", RV32I, opcode=0b0000011, funct3=0b000)
class InstructionLB(InstructionILType):
    def execute(self, model: Model):
        data = model.state.memory.lb((model.state.intreg[self.rs1] + self.imm).unsigned())
        if (data >> 7) & 0x1:
            data |= 0xFFFFFF00
        model.state.intreg[self.rd] = data


@isa("lh", RV32I, opcode=0b0000011, funct3=0b001)
class InstructionLH(InstructionILType):
    def execute(self, model: Model):
        data = model.state.memory.lh((model.state.intreg[self.rs1] + self.imm).unsigned())
        if (data >> 15) & 0x1:
            data |= 0xFFFF0000
        model.state.intreg[self.rd] = data


@isa("lw", RV32I, opcode=0b0000011, funct3=0b010)
class InstructionLW(InstructionILType):
    def execute(self, model: Model):
        data = model.state.memory.lw((model.state.intreg[self.rs1] + self.imm).unsigned())
        model.state.intreg[self.rd] = data


@isa("lbu", RV32I, opcode=0b0000011, funct3=0b100)
class InstructionLBU(InstructionILType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.memory.lb(
            (model.state.intreg[self.rs1] + self.imm).unsigned())


@isa("lhu", RV32I, opcode=0b0000011, funct3=0b101)
class InstructionLHU(InstructionILType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.memory.lh(
            (model.state.intreg[self.rs1] + self.imm).unsigned())


@isa("sb", RV32I, opcode=0b0100011, funct3=0b000)
class InstructionSB(InstructionSType):
    def execute(self, model: Model):
        model.state.memory.sb((model.state.intreg[self.rs1] + self.imm).unsigned(),
                        model.state.intreg[self.rs2])


@isa("sh", RV32I, opcode=0b0100011, funct3=0b001)
class InstructionSH(InstructionSType):
    def execute(self, model: Model):
        model.state.memory.sh((model.state.intreg[self.rs1] + self.imm).unsigned(),
                        model.state.intreg[self.rs2])


@isa("sw", RV32I, opcode=0b0100011, funct3=0b010)
class InstructionSW(InstructionSType):
    def execute(self, model: Model):
        model.state.memory.sw((model.state.intreg[self.rs1] + self.imm).unsigned(),
                        model.state.intreg[self.rs2])


@isa("addi", RV32I, opcode=0b0010011, funct3=0b000)
class InstructionADDI(InstructionIType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] + self.imm


@isa("slti", RV32I, opcode=0b0010011, funct3=0b010)
class InstructionSLTI(InstructionIType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1] < self.imm:
            model.state.intreg[self.rd] = 1
        else:
            model.state.intreg[self.rd] = 0


@isa("sltiu", RV32I, opcode=0b0010011, funct3=0b011)
class InstructionSLTIU(InstructionIType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1].unsigned() < int(self.imm):
            model.state.intreg[self.rd] = 1
        else:
            model.state.intreg[self.rd] = 0


@isa("xori", RV32I, opcode=0b0010011, funct3=0b100)
class InstructionXORI(InstructionIType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] ^ self.imm


@isa("ori", RV32I, opcode=0b0010011, funct3=0b110)
class InstructionORI(InstructionIType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] | self.imm


@isa("andi", RV32I, opcode=0b0010011, funct3=0b111)
class InstructionANDI(InstructionIType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] & self.imm


@isa("slli", RV32I, opcode=0b0010011, funct3=0b001, funct7=0b0000000)
class InstructionSLLI(InstructionISType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] << self.shamt


@isa("srli", RV32I, opcode=0b0010011, funct3=0b101, funct7=0b0000000)
class InstructionSRLI(InstructionISType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1].unsigned() >> int(
            self.shamt)


@isa("srai", RV32I, opcode=0b0010011, funct3=0b101, funct7=0b0100000)
class InstructionSRAI(InstructionISType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] >> self.shamt


@isa("add", RV32I, opcode=0b0110011, funct3=0b000, funct7=0b0000000)
class InstructionADD(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] + model.state.intreg[self.rs2]


@isa("sub", RV32I, opcode=0b0110011, funct3=0b000, funct7=0b0100000)
class InstructionSUB(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] - model.state.intreg[self.rs2]


@isa("sll", RV32I, opcode=0b0110011, funct3=0b001, funct7=0b0000000)
class InstructionSLL(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] << (
            model.state.intreg[self.rs2] & 0x1f)


@isa("slt", RV32I, opcode=0b0110011, funct3=0b010, funct7=0b0000000)
class InstructionSLT(InstructionRType):
    def execute(self, model: Model):
        if model.state.intreg[self.rs1] < model.state.intreg[self.rs2]:
            model.state.intreg[self.rd] = 1
        else:
            model.state.intreg[self.rd] = 0


@isa("sltu", RV32I, opcode=0b0110011, funct3=0b011, funct7=0b0000000)
class InstructionSLTU(InstructionRType):
    def execute(self, state: State):
        if state.intreg[self.rs1].unsigned() < state.intreg[
                self.rs2].unsigned():
            state.intreg[self.rd] = 1
        else:
            state.intreg[self.rd] = 0


@isa("xor", RV32I, opcode=0b0110011, funct3=0b100, funct7=0b0000000)
class InstructionXOR(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] ^ model.state.intreg[self.rs2]


@isa("srl", RV32I, opcode=0b0110011, funct3=0b101, funct7=0b0000000)
class InstructionSRL(InstructionRType):
    def execute(self, model: Model):
        src = model.state.intreg[self.rs1]
        shift = model.state.intreg[self.rs2] & 0x1f
        model.state.intreg[self.rd] = src >> shift


@isa("sra", RV32I, opcode=0b0110011, funct3=0b101, funct7=0b0100000)
class InstructionSRA(InstructionRType):
    def execute(self, model: Model):
        usrc = model.state.intreg[self.rs1].unsigned()
        shift = model.state.intreg[self.rs2].unsigned() & 0x1f
        if usrc >> 31:
            to_clear = 32 - shift
            sign_mask = (((1 << 32) - 1) >> to_clear) << to_clear
        else:
            sign_mask = 0

        model.state.intreg[self.rd] = sign_mask | (usrc >> shift)


@isa("or", RV32I, opcode=0b0110011, funct3=0b110, funct7=0b0000000)
class InstructionOR(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] | model.state.intreg[self.rs2]


@isa("and", RV32I, opcode=0b0110011, funct3=0b111, funct7=0b0000000)
class InstructionAND(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] & model.state.intreg[self.rs2]


@isa("fence", RV32I, opcode=0b0001111, funct3=0b000)
class InstructionFENCE(InstructionIType):
    isa_format_id = "FENCE"

    def execute(self, model: Model):
        pass


@isa("fence.i", RV32IZifencei, opcode=0b0001111, funct3=0b001)
class InstructionFENCEI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("ecall", RV32I, opcode=0b1110011, funct3=0b000, imm=0b000000000000, rd=0b00000, rs1=0b00000)
class InstructionECALL(InstructionIType):
    def execute(self, model: Model):
        model.environment.call(model.state)

    def __str__(self):
        return "ecall"


@isa("uret", RV32I, opcode=0b1110011, funct3=0b000, imm=0b000000000010, rs1=0b00000, rd=0b00000)
class InstructionURET(InstructionIType):
    """ Machine level exception return """
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("sret", RV32I, opcode=0b1110011, funct3=0b000, imm=0b000100000010, rs1=0b00000, rd=0b00000)
class InstructionSRET(InstructionIType):
    """ Machine level exception return """
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("hret", RV32I, opcode=0b1110011, funct3=0b000, imm=0b001000000010, rs1=0b00000, rd=0b00000)
class InstructionHRET(InstructionIType):
    """ Machine level exception return """
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("mret", RV32I, opcode=0b1110011, funct3=0b000, imm=0b001100000010, rs1=0b00000, rd=0b00000)
class InstructionMRET(InstructionIType):
    """ Machine level exception return """
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("wfi", RV32I, opcode=0b1110011, funct3=0b000, imm=0b000100000101, rs1=0b00000, rd=0b00000)
class InstructionWFI(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("ebreak", RV32I, opcode=0b1110011, funct3=0b000, imm=0b000000000001)
class InstructionEBREAK(InstructionIType):
    def execute(self, model: Model):
        pass

    def __str__(self):
        return "ebreak"


@isa("csrrw", RV32IZicsr, opcode=0b1110011, funct3=0b001)
class InstructionCSRRW(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("csrrs", RV32IZicsr, opcode=0b1110011, funct3=0b010)
class InstructionCSRRS(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("csrrc", RV32IZicsr, opcode=0b1110011, funct3=0b011)
class InstructionCSRRC(InstructionIType):
    def execute(self, model: Model):
        pass


#@isa("csrrwi", RV32IZicsr, opcode=0b1110011, funct3=0b101)
#class InstructionCSRRWI(Instruction):
#    def execute(self, model: Model):
#        pass


#@isa("csrrsi", RV32IZicsr, opcode=0b1110011, funct3=0b110)
#class InstructionCSRRSI(Instruction):
#    def execute(self, model: Model):
#        pass


#@isa("csrrci", RV32IZicsr, opcode=0b1110011, funct3=0b111)
#class InstructionCSRRCI(Instruction):
#    def execute(self, model: Model):
#        pass


@isa("lwu", RV64I, opcode=0b0000011, funct3=0b110)
class InstructionLWU(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("ld", RV64I, opcode=0b0000011, funct3=0b011)
class InstructionLD(InstructionIType):
    def execute(self, model: Model):
        pass


@isa("sd", RV64I, opcode=0b0100011, funct3=0b011)
class InstructionSD(InstructionISType):
    def execute(self, model: Model):
        pass


@isa_pseudo()
class InstructionNOP(InstructionADDI):
    def __init__(self):
        super().__init__(0, 0, 0)

    def __str__(self):
        return "nop"


@isa("mul", RV32IM, opcode=0b0110011, funct3=0b000, funct7=0b0000001)
class InstructionMUL(InstructionRType):
    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs1] * model.state.intreg[self.rs2]


@isa("mulh", RV32IM, opcode=0b0110011, funct3=0b001, funct7=0b0000001)
class InstructionMULH(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("mulhsu", RV32IM, opcode=0b0110011, funct3=0b010, funct7=0b0000001)
class InstructionMULHSU(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("mulhu", RV32IM, opcode=0b0110011, funct3=0b011, funct7=0b0000001)
class InstructionMULHU(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("div", RV32IM, opcode=0b0110011, funct3=0b100, funct7=0b0000001)
class InstructionDIV(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("divu", RV32IM, opcode=0b0110011, funct3=0b101, funct7=0b0000001)
class InstructionDIVU(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("rem", RV32IM, opcode=0b0110011, funct3=0b110, funct7=0b0000001)
class InstructionREM(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa("remu", RV32IM, opcode=0b0110011, funct3=0b111, funct7=0b0000001)
class InstructionREMU(InstructionRType):
    def execute(self, model: Model):
        # TODO: implement
        pass


@isa_c("c.addi", RV32IC, opcode=1, funct3=0b000)
class InstructionCADDI(InstructionCIType):
    def expand(self):
        pass

    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rd] + self.imm


@isa_c("c.andi", RV32IC, opcode=1, funct3=0b100)
class InstructionCANDI(InstructionCBType):
    def expand(self):
        pass

    def execute(self, model: Model):
        pass


@isa_c("c.swsp", RV32IC, opcode=2, funct3=6)
class InstructionCSWSP(InstructionCSSType):
    def expand(self):
        pass

    def decode(self, machinecode: int):
        self.rs = (machinecode >> 2) & 0x1f
        imm12to9 = (machinecode >> 9) & 0xf
        imm8to7 = (machinecode >> 7) & 0x3
        self.imm.set_from_bits((imm8to7 << 4) | imm12to9)

    def execute(self, model: Model):
        pass


@isa_c("c.li", RV32IC, opcode=1, funct3=2)
class InstructionCLI(InstructionCIType):
    def expand(self):
        pass

    def execute(self, model: Model):
        model.state.intreg[self.rd] = self.imm


@isa_c("c.mv", RV32IC, opcode=2, funct4=8)
class InstructionCMV(InstructionCRType):
    def expand(self):
        pass

    def execute(self, model: Model):
        model.state.intreg[self.rd] = model.state.intreg[self.rs]


@isa("lr", RV32A, opcode=0b0101111, funct5=0b00010, funct3=0b010)
class InstructionLR(InstructionAMOType):
    """ Load reserved """
    def execute(self, model: Model):
        # Perform a normal load
        data = model.state.memory.lw(model.state.intreg[self.rs1].unsigned())
        model.state.intreg[self.rd] = data
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("sc", RV32A, opcode=0b0101111, funct5=0b00011, funct3=0b010)
class InstructionSC(InstructionAMOType):
    """ Store conditional """
    def execute(self, model: Model):
        # Check if this address is reserved
        if model.state.atomic_reserved(model.state.intreg[self.rs1]):
            model.state.memory.sw(
                model.state.intreg[self.rs1].unsigned(),
                model.state.intreg[self.rs2]
            )
            model.state.intreg[self.rd] = 0
        else:
            model.state.intreg[self.rd] = 1
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amoadd", RV32A, opcode=0b0101111, funct5=0b00000, funct3=0b010)
class InstructionAMOADD(InstructionAMOType):
    """ Atomic add operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            (model.state.intreg[self.rs2] + model.state.intreg[self.rd])
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amoxor", RV32A, opcode=0b0101111, funct5=0b00100, funct3=0b010)
class InstructionAMOXOR(InstructionAMOType):
    """ Atomic XOR operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            (model.state.intreg[self.rs2] ^ model.state.intreg[self.rd])
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amoor",   RV32A, opcode=0b0101111, funct5=0b01000, funct3=0b010)
class InstructionAMOOR(InstructionAMOType):
    """ Atomic OR operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            (model.state.intreg[self.rs2] | model.state.intreg[self.rd])
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amoand", RV32A, opcode=0b0101111, funct5=0b01100, funct3=0b010)
class InstructionAMOAND(InstructionAMOType):
    """ Atomic AND operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            (model.state.intreg[self.rs2] & model.state.intreg[self.rd])
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amomin", RV32A, opcode=0b0101111, funct5=0b10000, funct3=0b010)
class InstructionAMOMIN(InstructionAMOType):
    """ Atomic minimum operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            min(model.state.intreg[self.rs2], model.state.intreg[self.rd])
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amomax", RV32A, opcode=0b0101111, funct5=0b10100, funct3=0b010)
class InstructionAMOMAX(InstructionAMOType):
    """ Atomic maximum operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            max(model.state.intreg[self.rs2], model.state.intreg[self.rd])
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amominu", RV32A, opcode=0b0101111, funct5=0b11000, funct3=0b010)
class InstructionAMOMINU(InstructionAMOType):
    """ Atomic unsigned minimum operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            min(
                model.state.intreg[self.rs2].unsigned(),
                model.state.intreg[self.rd].unsigned()
            )
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amomaxu", RV32A, opcode=0b0101111, funct5=0b11100, funct3=0b010)
class InstructionAMOMAXU(InstructionAMOType):
    """ Atomic unsigned maximum operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            max(
                model.state.intreg[self.rs2].unsigned(),
                model.state.intreg[self.rd].unsigned()
            )
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])


@isa("amoswap", RV32A, opcode=0b0101111, funct5=0b00001, funct3=0b010)
class InstructionAMOSWAP(InstructionAMOType):
    """ Atomic swap operation """
    def execute(self, model: Model):
        # This models a single HART with 1 stage pipeline, so will always succeed
        model.state.intreg[self.rd] = model.state.memory.lw(
            model.state.intreg[self.rs1].unsigned()
        )
        model.state.memory.sw(
            model.state.intreg[self.rs1].unsigned(),
            model.state.intreg[self.rs2]
        )
        # Perform correct lock or release actions
        if self.rl: model.state.atomic_release(model.state.intreg[self.rs1])
        elif self.aq: model.state.atomic_acquire(model.state.intreg[self.rs1])



from riscvmodel.isa import *
from riscvmodel.insn import *
from riscvmodel.regnames import *
from riscvmodel.program import Program


class LRSCTest(Program):
    """Basic test of LR & SC instructions"""
    def __init__(self):
        super().__init__([
            # Perform load-reserved and acquire lock
            InstructionLR(rs1=x0, rd=x1, aq=1, rl=0),
            # Perform store-conditional and release lock (should succeed)
            InstructionSC(rs1=x0, rs2=x0, rd=x2, aq=0, rl=1),
            # Perform second store-conditional to same address (should fail)
            InstructionSC(rs1=x0, rs2=x0, rd=x3, aq=0, rl=1),
        ])

    def expects(self) -> dict:
        return {
            x0: 0,
            x2: 0, # Due to successful SC (2nd instruction)
            x3: 1, # Due to failed SC (3rd instruction)
        }




# Copyright Stefan Wallentowitz
# Licensed under the MIT License, see LICENSE for details.
# SPDX-License-Identifier: MIT
"""
This is the basic ISA definition. It describes the instruction formats and base
classes. Actual instructions are implemented in the insn module.
"""

from random import randrange
from abc import ABCMeta, abstractmethod
from collections import namedtuple

from .variant import Variant
from .model import Model
from .types import Immediate, Register
from .variant import RV32I

Field = namedtuple("Field", ["name", "base", "size", "offset", "description", "static", "value"])
Field.__new__.__defaults__ = (None, None, None, 0, None, False, None)

class Instruction(metaclass=ABCMeta):
    """
    Base class for instructions

    This is the abstract base class for all instruction. They are derived via
    their instruction type.
    """

    # Class members starting with "field_" are defined by the ISA
    field_opcode = Field(name="opcode", base=0, size=7, description="", static=True)

    # Class members starting with "isa_" are defined by the ISA and describe and
    # identify an instruction
    mnemonic = None
    coding = None
    isa_variant = None

    isa_format_id = None

    asm_arg_signature = ""

    @classmethod
    def asm_signature(cls):
        signature = cls.mnemonic
        if len(cls.asm_arg_signature) > 0:
            signature += " " + cls.asm_arg_signature
        return signature

    @classmethod
    def extract_field(cls, field, word):
        fname = "field_{}".format(field)
        base = getattr(cls, fname).base
        size = getattr(cls, fname).size
        if not isinstance(base, list):
            base = [base]
            size = [size]
        off = 0
        value = 0
        for part in range(len(base)):
            value |= ((word >> base[part]) & (2**size[part] - 1)) << off
            off += size[part]
        return value << getattr(cls, fname).offset

    @classmethod
    def set_field(cls, field, word, value):
        fname = "field_{}".format(field)
        base = getattr(cls, fname).base
        size = getattr(cls, fname).size
        if not isinstance(base, list):
            base = [base]
            size = [size]
        off = getattr(cls, fname).offset
        for part in range(len(base)):
            word |= (((value >> off) & (2**size[part] - 1)) << base[part])
            off += size[part]
        return word

    @classmethod
    def get_fields(cls):
        return [getattr(cls, member) for member in dir(cls) if member.startswith("field_")]

    @classmethod
    def get_static_fields(cls):
        return [field for field in cls.get_fields() if field.static]

    @classmethod
    def get_isa_format(cls, *, asdict: bool=False):
        fields = [getattr(cls, attr) for attr in dir(cls) if attr.startswith("field_")]
        if asdict:
            fields = [field._asdict() for field in fields]
        return {"id": cls.isa_format_id, "fields": fields}

    @classmethod
    def match(cls, word: int):
        """Try to match a machine code to this instruction"""
        for field in cls.get_static_fields():
            if cls.extract_field(field.name, word) != field.value:
                return False
        return True

    def ops_from_string(self, ops: str):
        """
        Extract operands from string
        """
        self.ops_from_list(ops.split(","))

    def ops_from_list(self, ops: list):
        """
        Extract operands from list of string
        """

    def randomize(self, variant: Variant):
        """
        Randomize this instruction

        This function randomizes the instance of an instruction according to the
        given variant.

        :param variant: RISC-V ISA variant :return: nothing
        """

    @abstractmethod
    def execute(self, model: Model):
        """
        Execute this instruction

        Execute the instruction on the given model

        :param model: RISC-V core model
        :return: nothing
        """

    def decode(self, word: int):
        """
        Decode a machine code and configure this instruction from it.

        :param word: Machine code as 32-bit integer
        :type word: int
        """
        for field in self.get_fields():
            if field.static:
                assert self.extract_field(field.name, word) == field.value
            else:
                attr = getattr(self, field.name)
                value = self.extract_field(field.name, word)
                if isinstance(attr, Register):
                    attr.set(value)
                elif isinstance(attr, Immediate):
                    attr.set_from_bits(value)
                else:
                    setattr(self, field.name, value)

    def encode(self) -> int:
        """
        TODO: document
        """
        word = 0
        for field in self.get_fields():
            if field.value is not None:
                word = self.set_field(field.name, word, field.value)
            else:
                value = getattr(self, field.name)
                if isinstance(value, Immediate):
                    word = self.set_field(field.name, word, value.unsigned())
                elif isinstance(value, bool):
                    word = self.set_field(field.name, word, 1 if value else 0)
                else:
                    word = self.set_field(field.name, word, value)
        return word

    def __str__(self):
        """
        Generate assembler code

        Generate the assembler code for this instruction

        :return: Assembly string
        """
        return str(self.mnemonic)

    # pylint: disable=R0201,W0613
    def inopstr(self, model) -> str:
        """
        TODO: document
        """
        return ""

    # pylint: disable=R0201,W0613
    def outopstr(self, model) -> str:
        """
        TODO: document
        """
        return ""

    def __setattr__(self, key, value):
        if key in self.__dict__ and isinstance(self.__dict__[key], Immediate):
            raise AttributeError(
                "Instruction does not allow to overwrite immediates, use set() on them"
            )
        super().__setattr__(key, value)

    def __eq__(self, other):
        for field in self.get_fields():
            if field.static:
                if getattr(self, field) != getattr(other, field):
                    return False
            else:
                if getattr(self, field[6:]) != getattr(other, field[6:]):
                    return False
        return True

class InstructionFunct3Type(Instruction, metaclass=ABCMeta):
    field_funct3 = Field(name="funct3", base=12, size=3, description="", static=True)

class InstructionFunct5Type(Instruction, metaclass=ABCMeta):
    field_funct5 = Field(name="funct5", base=27, size=5, description="", static=True)

class InstructionFunct7Type(Instruction, metaclass=ABCMeta):
    field_funct7 = Field(name="funct7", base=25, size=7, description="", static=True)

class InstructionRType(InstructionFunct3Type, InstructionFunct7Type, metaclass=ABCMeta):
    """
    R-type instructions are 3-register instructions which use two source
    registers and write one output register.

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param rs2: Source register 2
    :type rs2: int
    """

    isa_format_id = "R"
    asm_arg_signature = "<rd>, <rs1>, <rs2>"

    field_rd = Field(name="rd", base=7, size=5, description="")
    field_rs1 = Field(name="rs1", base=15, size=5, description="")
    field_rs2 = Field(name="rs2", base=20, size=5, description="")

    def __init__(self, rd: int = None, rs1: int = None, rs2: int = None):
        super(InstructionRType, self).__init__()
        # pylint: disable=C0103
        self.rd = rd
        self.rs1 = rs1
        self.rs2 = rs2

    def ops_from_list(self, ops):
        (self.rd, self.rs1, self.rs2) = [int(op[1:]) for op in ops]

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.xlen)
        self.rs1 = randrange(0, variant.xlen)
        self.rs2 = randrange(0, variant.xlen)

    def inopstr(self, model):
        opstr = "{:>3}={}, ".format("x{}".format(self.rs1),
                                    model.state.intreg[self.rs1])
        opstr += "{:>3}={} ".format("x{}".format(self.rs2),
                                    model.state.intreg[self.rs2])
        return opstr

    def outopstr(self, model):
        return "{:>3}={} ".format("x{}".format(self.rd),
                                  model.state.intreg[self.rd])

    def __str__(self):
        return "{} x{}, x{}, x{}".format(self.mnemonic, self.rd, self.rs1,
                                         self.rs2)


class InstructionIType(InstructionFunct3Type, metaclass=ABCMeta):
    """
    I-type instructions are registers that use one source register and an
    immediate to produce a new value for the destination register.

    Two specialization exist for this class: :class:`InstructionILType` for load
    instructions and :class:`InstructionISType` for instructions that shift by
    an immediate value.

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param imm: 12-bit signed immediate
    :type imm: int
    """

    isa_format_id = "I"
    asm_arg_signature = "<rd>, <rs1>, <imm>"

    field_rd = Field(name="rd", base=7, size=5, description="")
    field_rs1 = Field(name="rs1", base=15, size=5, description="")
    field_imm = Field(name="imm", base=20, size=12, description="")

    def __init__(self, rd: int = None, rs1: int = None, imm: int = None):
        super(InstructionIType, self).__init__()
        self.rd = rd  # pylint: disable=C0103
        self.rs1 = rs1
        self.imm = Immediate(bits=12, signed=True)
        if imm is not None:
            self.imm.set(imm)

    def ops_from_list(self, ops):
        if len(ops) == 0: # ecall
            return
        self.rd = int(ops[0][1:])
        if ops[1][0] == "x":
            self.rs1 = int(ops[1][1:])
            self.imm.set(int(ops[2], 0))
        else: # Load
            self.rs1 = int(ops[2][1:])
            self.imm.set(int(ops[1], 0))

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.xlen)
        self.rs1 = randrange(0, variant.xlen)
        self.imm.randomize()

    def inopstr(self, model) -> str:
        return "{:>3}={} ".format("x{}".format(self.rs1),
                                  model.state.intreg[self.rs1])

    def outopstr(self, model) -> str:
        return "{:>3}={} ".format("x{}".format(self.rd),
                                  model.state.intreg[self.rd])

    def __str__(self) -> str:
        return "{} x{}, x{}, {}".format(self.mnemonic, self.rd, self.rs1,
                                        self.imm)

class InstructionILType(InstructionIType, metaclass=ABCMeta):
    """
    I-type instruction specialization for stores. The produce a different
    assembler than the base class

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param imm: 12-bit signed immediate
    :type rs2: int
    """
    def __str__(self):
        return "{} x{}, {}(x{})".format(self.mnemonic, self.rd, self.imm,
                                        self.rs1)


class InstructionISType(InstructionFunct3Type,InstructionFunct7Type, metaclass=ABCMeta):
    """
    Similar to R-Type instruction specialization for shifts by immediate.

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param imm: 12-bit signed immediate
    :type imm: int
    """

    isa_format_id = "IS"

    field_rd = Field(name="rd", base=7, size=5, description="")
    field_rs1 = Field(name="rs1", base=15, size=5, description="")
    field_shamt = Field(name="shamt", base=20, size=5, description="")

    asm_arg_signature = "<rd>, <rs1>, <shamt>"

    def __init__(self, rd: int = None, rs1: int = None, shamt: int = None):
        super(InstructionISType, self).__init__()
        self.rd = rd
        self.rs1 = rs1
        self.shamt = Immediate(bits=5, init=shamt)

    def ops_from_list(self, ops):
        self.rd = int(ops[0][1:])
        self.rs1 = int(ops[1][1:])
        self.shamt.set(int(ops[2], 0))

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.xlen)
        self.rs1 = randrange(0, variant.xlen)
        self.shamt.randomize()

    def inopstr(self, model):
        return "{:>3}={} ".format("x{}".format(self.rs1),
                                  model.state.intreg[self.rs1])

    def __str__(self):
        return "{} x{}, x{}, 0x{:02x}".format(self.mnemonic, self.rd, self.rs1,
                                              self.shamt)


class InstructionSType(InstructionFunct3Type, metaclass=ABCMeta):
    """
    S-type instructions are used for stores. They don't have a destination
    register, but two source registers.

    :param rs1: Source register for base address
    :type rs1: int
    :param rs2: Source register for data
    :type rs2: int
    :param imm: Offset of store, for calculation of address relative to rs1
    :type imm: int
    """

    isa_format_id = "S"

    field_rs1 = Field(name="rs1", base=15, size=5, description="")
    field_rs2 = Field(name="rs2", base=20, size=5, description="")
    field_imm = Field(name="imm", base=[7, 25], size=[5, 7], description="")

    asm_arg_signature = "<rs2>, <imm>(<rs1>)"

    def __init__(self, rs1: int = None, rs2: int = None, imm: int = None):
        super(InstructionSType, self).__init__()
        self.rs1 = rs1
        self.rs2 = rs2
        self.imm = Immediate(bits=12, signed=True, init=imm)

    def ops_from_list(self, ops):
        self.rs1 = int(ops[2][1:])
        self.rs2 = int(ops[0][1:])
        self.imm.set(int(ops[1], 0))

    def randomize(self, variant: Variant):
        self.rs1 = randrange(0, variant.xlen)
        self.rs2 = randrange(0, variant.xlen)
        self.imm.randomize()

    def inopstr(self, model):
        opstr = "{:>3}={}, ".format("x{}".format(self.rs1),
                                    model.state.intreg[self.rs1])
        opstr += "{:>3}={}".format("x{}".format(self.rs2),
                                   model.state.intreg[self.rs2])
        return opstr

    def __str__(self):
        return "{} x{}, {}(x{})".format(self.mnemonic, self.rs2, self.imm,
                                        self.rs1)

class InstructionBType(InstructionFunct3Type, metaclass=ABCMeta):
    """
    B-type instructions encode branches. Branches have two source registers that
    are compared. They then change the program counter by the immediate value.

    :param rs1: Source 1 for comparison
    :type rs1: int
    :param rs2: Source 2 for comparison
    :type rs2: int
    :param imm: Immediate for branch destination address calculation (13-bit, signed,
           16-bit aligned)
    :type imm: int
    """

    isa_format_id = "B"

    field_rs1 = Field(name="rs1", base=15, size=5, description="")
    field_rs2 = Field(name="rs2", base=20, size=5, description="")
    field_imm = Field(name="imm", base=[8, 25, 7, 31], size=[4, 6, 1, 1], offset=1, description="")

    asm_arg_signature = "<rs1>, <rs2>, <imm>"

    def __init__(self, rs1: int = None, rs2: int = None, imm: int = None):
        super(InstructionBType, self).__init__()
        self.rs1 = rs1
        self.rs2 = rs2
        self.imm = Immediate(bits=13, signed=True, lsb0=True, init=imm)

    def ops_from_list(self, ops):
        self.rs1 = int(ops[0][1:])
        self.rs2 = int(ops[1][1:])
        self.imm.set(int(ops[2], 0))

    def randomize(self, variant: Variant):
        self.rs1 = randrange(0, variant.xlen)
        self.rs2 = randrange(0, variant.xlen)
        self.imm.randomize()

    def inopstr(self, model):
        opstr = "{:>3}={}, ".format("x{}".format(self.rs1),
                                    model.state.intreg[self.rs1])
        opstr += "{:>3}={}".format("x{}".format(self.rs2),
                                   model.state.intreg[self.rs2])
        return opstr

    def __str__(self):
        return "{} x{}, x{}, .{:+}".format(self.mnemonic, self.rs1, self.rs2,
                                           self.imm)

class InstructionUType(Instruction, metaclass=ABCMeta):
    """
    U-type instructions are used for constant formation and set the upper bits of a register.

    :param rd: Destination register
    :type rd: int
    :param imm: Immediate (20-bit, unsigned)
    :type imm: int
    """

    field_rd = Field(name="rd", base=7, size=5, description="")
    field_imm = Field(name="imm", base=12, size=20, description="")

    def __init__(self, rd: int = None, imm: int = None):
        super(InstructionUType, self).__init__()
        self.rd = rd  # pylint: disable=invalid-name
        self.imm = Immediate(bits=20, init=imm)

    def ops_from_list(self, ops):
        self.rd = int(ops[0][1:])
        self.imm.set(int(ops[1], 0))

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.xlen)
        self.imm.randomize()

    def outopstr(self, model):
        return "{:>3}={} ".format("x{}".format(self.rd),
                                  model.state.intreg[self.rd])

    def __str__(self):
        return "{} x{}, {}".format(self.mnemonic, self.rd, self.imm)


class InstructionJType(Instruction, metaclass=ABCMeta):
    """
    J-type instruction are used for jump and link instructions.

    :param rd: Destination register
    :type rd: int
    :param imm: Immediate for the jump (21-bit, signed, 16-bit aligned)
    :type imm: int
    """

    field_rd = Field(name="rd", base=7, size=5, description="")
    field_imm = Field(name="imm", base=[21,20,12,31], size=[10,1,8,1], description="", offset=1)

    def __init__(self, rd: int = None, imm: int = None):
        super(InstructionJType, self).__init__()
        self.rd = rd  # pylint: disable=invalid-name
        self.imm = Immediate(bits=21, signed=True, lsb0=True)
        if imm is not None:
            self.imm.set(imm)

    def ops_from_list(self, ops):
        self.rd = int(ops[0][1:])
        self.imm.set(int(ops[1]))

    def randomize(self, variant: Variant):
        self.rd = randrange(0, variant.xlen)
        self.imm.randomize()

    def outopstr(self, model):
        return "{:>3}={} ".format("x{}".format(self.rd),
                                  model.state.intreg[self.rd])

    def __str__(self):
        return "{} x{}, .{:+}".format(self.mnemonic, self.rd, self.imm)


class InstructionCType(Instruction, metaclass=ABCMeta):
    """
    Compact instructions
    """
    @abstractmethod
    def expand(self):
        """
        Expand to full instruction
        """


class InstructionCBType(InstructionCType, metaclass=ABCMeta):
    """
    TODO: document
    """
    def __init__(self, rd: int = None, imm: int = None):
        super(InstructionCBType, self).__init__()
        self.rd = rd  # pylint: disable=invalid-name
        if self.rd is not None:
            self.rd = rd + 8
        self.imm = Immediate(bits=6, signed=True, lsb0=True)
        if imm is not None:
            self.imm.set(imm)

    def decode(self, machinecode: int):
        self.rd = machinecode

    def __str__(self):
        return "{} x{}, {}".format(self.mnemonic, self.rd, self.imm)


class InstructionCRType(InstructionCType, metaclass=ABCMeta):
    """
    TODO: document
    """
    def __init__(self, rd: int = None, rs: int = None):
        super(InstructionCRType, self).__init__()
        self.rd = rd  # pylint: disable=invalid-name
        self.rs = rs  # pylint: disable=invalid-name

    def randomize(self, variant: Variant):
        self.rd = randrange(8, 16)
        self.rs = randrange(8, 16)

    def __str__(self):
        return "{} x{}, x{}".format(self.mnemonic, self.rd, self.rs)


class InstructionCIType(InstructionCType, metaclass=ABCMeta):
    """
    TODO: document
    """
    def __init__(self, rd: int = None, imm: int = None):
        super(InstructionCIType, self).__init__()
        self.rd = rd  # pylint: disable=invalid-name
        self.imm = Immediate(bits=6, signed=True, lsb0=True)
        if imm is not None:
            self.imm.set(imm)

    def randomize(self, variant: Variant):
        self.rd = randrange(0, 16)
        self.imm.randomize()

    def decode(self, machinecode: int):
        self.rd = (machinecode >> 7) & 0x1F
        imm12 = (machinecode >> 12) & 0x1
        imm6to2 = (machinecode >> 2) & 0x1F
        self.imm.set_from_bits((imm12 << 5) | imm6to2)

    def __str__(self):
        return "{} x{}, {}".format(self.mnemonic, self.rd, self.imm)


class InstructionCSSType(InstructionCType, metaclass=ABCMeta):
    """
    TODO: document
    """
    def __init__(self, rs: int = None, imm: int = None):
        super(InstructionCSSType, self).__init__()
        self.rs = rs  # pylint: disable=invalid-name
        self.imm = Immediate(bits=6, signed=True, lsb0=True)
        if imm is not None:
            self.imm.set(imm)

    def __str__(self):
        return "{} x{}, {}(x2)".format(self.mnemonic, self.rs, self.imm)

    def randomize(self, variant: Variant):
        self.rs = randrange(0, 16)
        self.imm.randomize()


class InstructionAMOType(InstructionFunct3Type, InstructionFunct5Type, metaclass=ABCMeta):
    """
    AMO-type instructions used a modified version of the R-type instruction.
    These are also 3-register instructions (use 2 source, write 1 output) but
    have additional flags for controlling acquisition (aq) and release (rl) of
    locks on target addresses.

    :param rd: Destination register
    :type rd: int
    :param rs1: Source register 1
    :type rs1: int
    :param rs2: Source register 2
    :type rs2: int
    :param rl: Lock release flag
    :type rl: int
    :param aq: Lock acquisition flag
    :type aq: int
    """

    isa_format_id     = "R"
    asm_arg_signature = "<rd>, <rs1>, <rs2>, <rl>, <aq>"

    field_rd  = Field(name="rd", base=7, size=5, description="")
    field_rs1 = Field(name="rs1", base=15, size=5, description="")
    field_rs2 = Field(name="rs2", base=20, size=5, description="")
    field_rl  = Field(name="rl",  base=25, size=1, description="Lock release")
    field_aq  = Field(name="aq",  base=26, size=1, description="Lock acquire")

    def __init__(
        self,
        rd: int = None,
        rs1: int = None,
        rs2: int = None,
        rl: int = None,
        aq: int = None,
    ):
        super(InstructionAMOType, self).__init__()
        self.rd  = rd
        self.rs1 = rs1
        self.rs2 = rs2
        self.rl  = rl
        self.aq  = aq

    def ops_from_list(self, ops):
        (self.rd, self.rs1, self.rs2, self.rl, self.aq) = [int(op[1:]) for ops in ops]

    def randomize(self, variant: Variant):
        self.rd  = randrange(0, variant.xlen)
        self.rs1 = randrange(0, variant.xlen)
        self.rs2 = randrange(0, variant.xlen)
        self.rl  = randrange(0, variant.xlen)
        self.aq  = randrange(0, variant.xlen)

    def inopstr(self, model):
        opstr = "{:>3}={}, ".format(
            "x{}".format(self.rs1), model.state.intreg[self.rs1]
        )
        opstr += "{:>3}={} ".format(
            "x{}".format(self.rs2), model.state.intreg[self.rs2]
        )
        return opstr

    def outopstr(self, model):
        return "{:>3}={} ".format(
            "x{}".format(self.rd), model.state.intreg[self.rd]
        )

    def __str__(self):
        return "{} x{}, x{}, x{}, {}, {}".format(
            self.mnemonic, self.rd, self.rs1, self.rs2, self.rl, self.aq
        )


def isa(mnemonic: str,
        variant: Variant,
        *,
        opcode: int,
        **kwargs
        ):
    """
    Decorator for the instructions. The decorator contains the static information for the
    instructions, in particular the encoding parameters and the assembler mnemonic.

    :param mnemonic: Assembler mnemonic
    :param opcode: Opcode of this instruction
    :return: Wrapper class that overwrites the actual definition and contains static data
    """
    def wrapper(wrapped):
        wrapped.field_opcode = wrapped.field_opcode._replace(value=opcode)

        wrapped.mnemonic = mnemonic
        wrapped.variant = variant

        for field in kwargs:
            fid = "field_"+field
            assert fid in dir(wrapped), "Invalid field {} for {}".format(fid, wrapped.__name__)
            setattr(wrapped, fid, getattr(wrapped, fid)._replace(value=kwargs[field]))

        return wrapped
    return wrapper


def isa_c(mnemonic: str,
          variant: Variant,
          *,
          opcode: int,
          funct3=None,
          funct4=None,
          funct6=None):
    """
    Decorator for the instructions. The decorator contains the static information for the
    instructions, in particular the encoding parameters and the assembler mnemonic.

    :param mnemonic: Assembler mnemonic
    :return: Wrapper class that overwrites the actual definition and contains static data
    """
    def wrapper(wrapped):
        """Get wrapper"""
        class WrappedClass(wrapped):  # pylint: disable=too-few-public-methods
            """Generic wrapper class"""

        WrappedClass.__name__ = wrapped.__name__
        WrappedClass.__module__ = wrapped.__module__
        WrappedClass.__qualname__ = wrapped.__qualname__
        WrappedClass.__doc__ = wrapped.__doc__
        return WrappedClass

    return wrapper


def isa_pseudo():
    """
    TODO: documentation
    """
    def wrapper(wrapped):
        class WrappedClass(wrapped):  # pylint: disable=too-few-public-methods
            """
            Wrapper class
            """

            _pseudo = True

        WrappedClass.__name__ = wrapped.__name__
        WrappedClass.__module__ = wrapped.__module__
        WrappedClass.__qualname__ = wrapped.__qualname__
        WrappedClass.__doc__ = wrapped.__doc__
        return WrappedClass

    return wrapper


def get_insns(*, cls=None, variant: Variant = RV32I):
    """
    Get all Instructions. This is based on all known subclasses of `cls`. If non
    is given, all Instructions are returned. Only such instructions are returned
    that can be generated, i.e., that have a mnemonic, opcode, etc. So other
    classes in the hierarchy are not matched.

    :param cls: Base class to get list :type cls: Instruction :return: List of
    instruction classes
    """
    insns = []

    if cls is None:
        cls = Instruction

    # This filters out abstract classes
    if cls.mnemonic:
        if variant is None or cls.variant <= variant:
            insns = [cls]

    for subcls in cls.__subclasses__():
        insns += get_insns(cls=subcls, variant=variant)

    insns = list(dict.fromkeys(insns)) # Remove duplicates
    return insns


def reverse_lookup(mnemonic: str, variant: Variant = None):
    """
    Find instruction that matches the mnemonic.

    :param mnemonic: Mnemonic to match
    :return: :class:`Instruction` that matches or None
    """
    for i in get_insns(variant=variant):
        if i.mnemonic == mnemonic:
            return i

    return None


def get_mnemomics():
    """
    Get all known mnemonics

    :return: List of all known mnemonics
    :rtype: List[str]
    """
    return [i.mnemonic for i in get_insns()]




from riscvmodel.isa import *
from riscvmodel.insn import *
from riscvmodel.regnames import *
from riscvmodel.program import Program


class LUITest(Program):
    """Basic test of LUI instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0),
            InstructionLUI(x2, 1),
            InstructionLUI(x3, 0x80000),
            InstructionLUI(x4, 0xfffff),
            InstructionLUI(x0, 0xfffff),
        ]
        super().__init__(insns)
    def expects(self) -> dict:
        return {0: 0, 1: 0, 2: 1, 3: 0x80000000, 4: 0xfffff000}


class AUIPCTest(Program):
    """Basic test of AUIPC instruction"""
    def __init__(self):
        insns = [
            InstructionAUIPC(x1, 0),
            InstructionAUIPC(x2, 1),
            InstructionAUIPC(x3, 0x80000),
            InstructionAUIPC(x4, 0xfffff),
            InstructionAUIPC(x0, 0xfffff),
        ]
        super().__init__(insns)


class ADDITest(Program):
    """Basic test of ADDI instruction"""
    def __init__(self):
        insns = [
            InstructionADDI(x1, x0, x0),
            InstructionADDI(x1, x1, 1),
            InstructionADDI(x2, x1, -2),
        ]
        super().__init__(insns)

    def expects(self) -> dict:
        return {1: 1, 2: 0xFFFFFFFF}


class SLTITest(Program):
    """Basic test of SLTI instruction"""
    def __init__(self):
        insns = [
            InstructionSLTI(x1, x0, 0),  # x1=0
            InstructionSLTI(x2, x1, 0),  # x2=0
            InstructionSLTI(x3, x0, 1),  # x3=1
            InstructionSLTI(x4, x3, 1),  # x4=0
            InstructionLUI(x5, 0x80000),
            InstructionSLTI(x6, x5, 0),  # x6=1
        ]
        super().__init__(insns)


class SLTIUTest(Program):
    """Basic test of SLTIU instruction"""
    def __init__(self):
        insns = [
            InstructionSLTIU(x1, x0, 0),
            InstructionSLTIU(x2, x1, 0),
            InstructionSLTIU(x3, x0, 1),
            InstructionSLTIU(x4, x3, 1),
            InstructionLUI(x5, 0x80000),
            InstructionSLTIU(x6, x5, 0),
        ]
        super().__init__(insns)


class XORITest(Program):
    """Basic test of XORI instruction"""
    def __init__(self):
        insns = [
            InstructionXORI(x1, x0, 0x7ff),
            InstructionXORI(x2, x1, 0x001),
            InstructionXORI(x3, x0, -0x800),
        ]
        super().__init__(insns)


class ORITest(Program):
    """Basic test of ORI instruction"""
    def __init__(self):
        insns = [
            InstructionORI(x1, x0, 0x7ff),
            InstructionORI(x2, x1, 0x001),
            InstructionORI(x3, x0, -0x800),
        ]
        super().__init__(insns)


class ANDITest(Program):
    """Basic test of ANDI instruction"""
    def __init__(self):
        insns = [
            InstructionANDI(x1, x0, 0x7ff),
            InstructionANDI(x2, x1, 0x001),
            InstructionANDI(x3, x0, -0x800),
        ]
        super().__init__(insns)


class SLLITest(Program):
    """Basic test of SLLI instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0xfebed),
            InstructionSLLI(x2, x1, 12),
            InstructionSLLI(x3, x1, 2),
            InstructionSLLI(x4, x3, 10),
        ]
        super().__init__(insns)


class SRLITest(Program):
    """Basic test of SRLI instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0xfebed),
            InstructionSRLI(x2, x1, 12),
            InstructionSRLI(x3, x1, 2),
            InstructionSRLI(x4, x3, 10),
        ]
        super().__init__(insns)


class SRAITest(Program):
    """Basic test of SRAI instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0xfebed),
            InstructionSRAI(x2, x1, 12),
            InstructionSRAI(x3, x1, 2),
            InstructionSRAI(x4, x3, 10),
        ]
        super().__init__(insns)


class ADDTest(Program):
    """Basic test of ADD instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0xfebed),
            InstructionADD(x2, x0, x1),
            InstructionADD(x3, x1, x1),
        ]
        super().__init__(insns)


class SUBTest(Program):
    """Basic test of SUB instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0xfebed),
            InstructionSUB(x2, x0, x1),
            InstructionSUB(x3, x1, x1),
        ]
        super().__init__(insns)


class SLLTest(Program):
    """Basic test of SLL instruction"""
    def __init__(self):
        insns = [
            InstructionLUI(x1, 0xfebed),
            InstructionSLL(x2, x1, x1),
            InstructionORI(x3, x0, 7),
            InstructionSLL(x4, x1, x3),
        ]
        super().__init__(insns)


class SLTTest(Program):
    """Basic test of SLT instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


class SLTUTest(Program):
    """Basic test of SLTU instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


class XORTest(Program):
    """Basic test of XOR instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


class SRLTest(Program):
    """Basic test of SRL instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


class SRATest(Program):
    """Basic test of SRA instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


class ORTest(Program):
    """Basic test of OR instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


class ANDTest(Program):
    """Basic test of AND instruction"""
    def __init__(self):
        insns = [
        ]
        super().__init__(insns)


RV32ITests = [LUITest, AUIPCTest, ADDITest, SLTITest, SLTIUTest, XORITest, ORITest, ANDITest, SLLITest, SRLITest,
              SRAITest, ADDTest, SUBTest, SLLTest, SLTTest, SLTUTest, XORTest, SRLTest, SRATest, ORTest, ANDTest]


# Ported from https://github.com/riscv/riscv-binutils-gdb

# Privileged CSR addresses (v1.11).
ustatus        = 0x0
uie            = 0x4
utvec          = 0x5
uscratch       = 0x40
uepc           = 0x41
ucause         = 0x42
utval          = 0x43
uip            = 0x44
cycle          = 0xc00
time           = 0xc01
instret        = 0xc02
hpmcounter3    = 0xc03
hpmcounter4    = 0xc04
hpmcounter5    = 0xc05
hpmcounter6    = 0xc06
hpmcounter7    = 0xc07
hpmcounter8    = 0xc08
hpmcounter9    = 0xc09
hpmcounter10   = 0xc0a
hpmcounter11   = 0xc0b
hpmcounter12   = 0xc0c
hpmcounter13   = 0xc0d
hpmcounter14   = 0xc0e
hpmcounter15   = 0xc0f
hpmcounter16   = 0xc10
hpmcounter17   = 0xc11
hpmcounter18   = 0xc12
hpmcounter19   = 0xc13
hpmcounter20   = 0xc14
hpmcounter21   = 0xc15
hpmcounter22   = 0xc16
hpmcounter23   = 0xc17
hpmcounter24   = 0xc18
hpmcounter25   = 0xc19
hpmcounter26   = 0xc1a
hpmcounter27   = 0xc1b
hpmcounter28   = 0xc1c
hpmcounter29   = 0xc1d
hpmcounter30   = 0xc1e
hpmcounter31   = 0xc1f
cycleh         = 0xc80
timeh          = 0xc81
instreth       = 0xc82
hpmcounter3h   = 0xc83
hpmcounter4h   = 0xc84
hpmcounter5h   = 0xc85
hpmcounter6h   = 0xc86
hpmcounter7h   = 0xc87
hpmcounter8h   = 0xc88
hpmcounter9h   = 0xc89
hpmcounter10h  = 0xc8a
hpmcounter11h  = 0xc8b
hpmcounter12h  = 0xc8c
hpmcounter13h  = 0xc8d
hpmcounter14h  = 0xc8e
hpmcounter15h  = 0xc8f
hpmcounter16h  = 0xc90
hpmcounter17h  = 0xc91
hpmcounter18h  = 0xc92
hpmcounter19h  = 0xc93
hpmcounter20h  = 0xc94
hpmcounter21h  = 0xc95
hpmcounter22h  = 0xc96
hpmcounter23h  = 0xc97
hpmcounter24h  = 0xc98
hpmcounter25h  = 0xc99
hpmcounter26h  = 0xc9a
hpmcounter27h  = 0xc9b
hpmcounter28h  = 0xc9c
hpmcounter29h  = 0xc9d
hpmcounter30h  = 0xc9e
hpmcounter31h  = 0xc9f
sstatus        = 0x100
sedeleg        = 0x102
sideleg        = 0x103
sie            = 0x104
stvec          = 0x105
scounteren     = 0x106
sscratch       = 0x140
sepc           = 0x141
scause         = 0x142
stval          = 0x143
sip            = 0x144
satp           = 0x180
mvendorid      = 0xf11
marchid        = 0xf12
mimpid         = 0xf13
mhartid        = 0xf14
mstatus        = 0x300
misa           = 0x301
medeleg        = 0x302
mideleg        = 0x303
mie            = 0x304
mtvec          = 0x305
mcounteren     = 0x306
mscratch       = 0x340
mepc           = 0x341
mcause         = 0x342
mtval          = 0x343
mip            = 0x344
pmpcfg0        = 0x3a0
pmpcfg1        = 0x3a1
pmpcfg2        = 0x3a2
pmpcfg3        = 0x3a3
pmpaddr0       = 0x3b0
pmpaddr1       = 0x3b1
pmpaddr2       = 0x3b2
pmpaddr3       = 0x3b3
pmpaddr4       = 0x3b4
pmpaddr5       = 0x3b5
pmpaddr6       = 0x3b6
pmpaddr7       = 0x3b7
pmpaddr8       = 0x3b8
pmpaddr9       = 0x3b9
pmpaddr10      = 0x3ba
pmpaddr11      = 0x3bb
pmpaddr12      = 0x3bc
pmpaddr13      = 0x3bd
pmpaddr14      = 0x3be
pmpaddr15      = 0x3bf
mcycle         = 0xb00
minstret       = 0xb02
mhpmcounter3   = 0xb03
mhpmcounter4   = 0xb04
mhpmcounter5   = 0xb05
mhpmcounter6   = 0xb06
mhpmcounter7   = 0xb07
mhpmcounter8   = 0xb08
mhpmcounter9   = 0xb09
mhpmcounter10  = 0xb0a
mhpmcounter11  = 0xb0b
mhpmcounter12  = 0xb0c
mhpmcounter13  = 0xb0d
mhpmcounter14  = 0xb0e
mhpmcounter15  = 0xb0f
mhpmcounter16  = 0xb10
mhpmcounter17  = 0xb11
mhpmcounter18  = 0xb12
mhpmcounter19  = 0xb13
mhpmcounter20  = 0xb14
mhpmcounter21  = 0xb15
mhpmcounter22  = 0xb16
mhpmcounter23  = 0xb17
mhpmcounter24  = 0xb18
mhpmcounter25  = 0xb19
mhpmcounter26  = 0xb1a
mhpmcounter27  = 0xb1b
mhpmcounter28  = 0xb1c
mhpmcounter29  = 0xb1d
mhpmcounter30  = 0xb1e
mhpmcounter31  = 0xb1f
mcycleh        = 0xb80
minstreth      = 0xb82
mhpmcounter3h  = 0xb83
mhpmcounter4h  = 0xb84
mhpmcounter5h  = 0xb85
mhpmcounter6h  = 0xb86
mhpmcounter7h  = 0xb87
mhpmcounter8h  = 0xb88
mhpmcounter9h  = 0xb89
mhpmcounter10h = 0xb8a
mhpmcounter11h = 0xb8b
mhpmcounter12h = 0xb8c
mhpmcounter13h = 0xb8d
mhpmcounter14h = 0xb8e
mhpmcounter15h = 0xb8f
mhpmcounter16h = 0xb90
mhpmcounter17h = 0xb91
mhpmcounter18h = 0xb92
mhpmcounter19h = 0xb93
mhpmcounter20h = 0xb94
mhpmcounter21h = 0xb95
mhpmcounter22h = 0xb96
mhpmcounter23h = 0xb97
mhpmcounter24h = 0xb98
mhpmcounter25h = 0xb99
mhpmcounter26h = 0xb9a
mhpmcounter27h = 0xb9b
mhpmcounter28h = 0xb9c
mhpmcounter29h = 0xb9d
mhpmcounter30h = 0xb9e
mhpmcounter31h = 0xb9f
mcountinhibit  = 0x320
mhpmevent3     = 0x323
mhpmevent4     = 0x324
mhpmevent5     = 0x325
mhpmevent6     = 0x326
mhpmevent7     = 0x327
mhpmevent8     = 0x328
mhpmevent9     = 0x329
mhpmevent10    = 0x32a
mhpmevent11    = 0x32b
mhpmevent12    = 0x32c
mhpmevent13    = 0x32d
mhpmevent14    = 0x32e
mhpmevent15    = 0x32f
mhpmevent16    = 0x330
mhpmevent17    = 0x331
mhpmevent18    = 0x332
mhpmevent19    = 0x333
mhpmevent20    = 0x334
mhpmevent21    = 0x335
mhpmevent22    = 0x336
mhpmevent23    = 0x337
mhpmevent24    = 0x338
mhpmevent25    = 0x339
mhpmevent26    = 0x33a
mhpmevent27    = 0x33b
mhpmevent28    = 0x33c
mhpmevent29    = 0x33d
mhpmevent30    = 0x33e
mhpmevent31    = 0x33f
hstatus        = 0x200
hedeleg        = 0x202
hideleg        = 0x203
hie            = 0x204
htvec          = 0x205
hscratch       = 0x240
hepc           = 0x241
hcause         = 0x242
hbadaddr       = 0x243
hip            = 0x244
mbase          = 0x380
mbound         = 0x381
mibase         = 0x382
mibound        = 0x383
mdbase         = 0x384
mdbound        = 0x385
mscounteren    = 0x321
mhcounteren    = 0x322

# Unprivileged CSR addresses
fflags         = 0x1
frm            = 0x2
fcsr           = 0x3
dcsr           = 0x7b0
dpc            = 0x7b1
dscratch0      = 0x7b2
dscratch1      = 0x7b3
tselect        = 0x7a0
tdata1         = 0x7a1
tdata2         = 0x7a2
tdata3         = 0x7a3
tinfo          = 0x7a4
tcontrol       = 0x7a5
mcontext       = 0x7a8
scontext       = 0x7aa


from random import randint
from enum import Enum
from collections import namedtuple


class InvalidImmediateException(Exception):
    """
    This exception is generated by Immediates for invalid values. It contains the message for the reason.
    """

class Immediate(object):
    """
    Immediate values are stored in this container, which safeguards them. An Immediate is configured for a bit width and
    can be signed or unsigned. Finally, there are immediates in RISC-V which are aligned to instruction address
    granularity, so that an immediate can be configured to be aligned to 16-bit boundaries (lsb = 0).

    :param bits: bit width of the immediate
    :type bits: int
    :param signed: Signedness of the immediate
    :type signed: bool
    :param lsb0: Set to True if this immediate is aligned to 16-bit boundaries
    :type lsb0: bool
    """
    def __init__(self, *, bits: int, signed: bool = False, lsb0: bool = False, init: int = None):
        self.bits = bits
        self.signed = signed
        self.lsb0 = lsb0
        self.value = 0
        self.tcmask = 1 << (self.bits - 1) # mask used for two's complement
        self.mask = (1 << self.bits) - 1
        if init is not None:
            self.set(init)

    def max(self) -> int:
        """
        Get the maximum value this immediate can have

        :return: Maximum value of this immediate
        """
        if self.signed:
            v = (1 << (self.bits - 1)) - 1
        else:
            v = (1 << self.bits) - 1
        if self.lsb0:
            v = v - (v % 2)
        return v

    def min(self) -> int:
        """
        Get the minimum value this immediate can have

        :return: Minimum value of this immediate
        """
        if self.signed:
            return -(1 << (self.bits - 1))
        else:
            return 0

    def exception(self, msg: str) -> InvalidImmediateException:
        # Generate exception
        message = "Immediate(bits={}, signed={}, lsb0={}) {}".format(self.bits, self.signed, self.lsb0, msg)
        return InvalidImmediateException(message)

    def set(self, value: int):
        """
        Set the immediate to a value. This function checks if the value is valid and will raise an
        :class:`InvalidImmediateException` if it doesn't.

        :param value: Value to set the immediate to
        :type value: int

        :raises InvalidImmediateException: value does not match immediate
        """
        if not isinstance(value, int):
            raise self.exception("{} is not an integer".format(value))
        if self.lsb0 and self.value % 2 == 1:
            raise self.exception("{} not power of two".format(value))
        if not self.signed and value < 0:
            raise self.exception("{} cannot be negative".format(value))
        if value < self.min() or value > self.max():
            raise self.exception("{} not in allowed range {}-{}".format(value, self.min(), self.max()))

        self.value = value

    def set_from_bits(self, value: int):
        """
        Set the immediate value from machine code bits. Those are not sign extended, so it will take care of the
        proper handling.

        :param value: Value to set the immediate to
        :type value: int
        """
        if self.signed:
            value = -(value & self.tcmask) + (value & ~self.tcmask)
        self.set(value)

    def randomize(self):
        """
        Randomize this immediate to a legal value
        """
        self.value = randint(self.min(), self.max())
        if self.lsb0:
            self.value = self.value - (self.value % 2)

    def __int__(self):
        """Convert to int"""
        return self.value.__int__()

    def unsigned(self):
        return self.value & self.mask

    def __str__(self):
        """Convert to string"""
        return self.value.__str__()

    def __repr__(self):
        return "Immediate(bits={}, signed={}, lsb0={}, value={:x})".format(self.bits, self.signed, self.lsb0, self.value)

    def __format__(self, format_spec):
        """Apply format spec"""
        return self.value.__format__(format_spec)

    def __lshift__(self, shamt):
        bits = self.bits + shamt
        new = Immediate(bits=bits, signed=self.signed, lsb0=self.lsb0)
        new.set(self.value << shamt)
        return new

    def __eq__(self, other):
        if isinstance(other, int):
            return self.value == other
        else:
            return self.value == other.value


class Register(object):
    def __init__(self, bits: int):
        self.bits = bits
        self.immutable = False
        self.value = 0
        self.format = "{{:0{}x}}".format(int(bits/4))
        self.mask = (1 << bits) - 1

    def set_immutable(self, s: bool):
        self.immutable = s

    def randomize(self):
        if not self.immutable:
            self.value = randint(0, 1 << self.bits - 1)

    def set(self, value):
        if isinstance(value, (Register, Immediate)):
            value = value.value
        if not self.immutable:
            self.value = value
        if (self.value >> (self.bits - 1)) & 1 != 0:
            self.value |= ~self.mask
        else:
            self.value &= self.mask

    def __int__(self):
        return self.value

    def unsigned(self):
        return self.value & self.mask

    def __str__(self):
        return self.format.format(self.value & self.mask)

    def __add__(self, other):
        new = Register(self.bits)
        new.set(self.value + int(other))
        return new

    def __sub__(self, other):
        new = Register(self.bits)
        new.set(self.value - int(other))
        return new

    # todo: doesn't work
    def __cmp__(self, other):
        return self.bits != other.bits or self.value != other.value

    def __and__(self, other):
        new = Register(self.bits)
        if isinstance(other, int):
            new.set(self.value & other)
        elif isinstance(other, (Register, Immediate)):
            new.set(self.value & other.value)
        else:
            raise TypeError("unsupported operand type for Register &: {}".format(other.__class__))
        return new

    def __or__(self, other):
        new = Register(self.bits)
        if isinstance(other, int):
            new.set(self.value | other)
        elif isinstance(other, (Register, Immediate)):
            new.set(self.value | other.value)
        else:
            raise TypeError("unsupported operand type for Register |: {}".format(other.__class__))
        return new

    def __lt__(self, other):
        return self.value < int(other)

    def __xor__(self, other):
        new = Register(self.bits)
        if isinstance(other, int):
            new.set(self.value ^ other)
        elif isinstance(other, (Register, Immediate)):
            new.set(self.value ^ other.value)
        else:
            raise TypeError("unsupported operand type for Register ^: {}".format(other.__class__))
        return new

    def __lshift__(self, other):
        new = Register(self.bits)
        if isinstance(other, int):
            new.set(self.value << other)
        elif isinstance(other, (Register, Immediate)):
            new.set(self.value << other.value)
        else:
            raise TypeError("unsupported operand type for Register <<: {}".format(other.__class__))
        return new

    def __rshift__(self, other):
        new = Register(self.bits)
        value = (self.value + 2**self.bits) if self.value < 0 else self.value
        if isinstance(other, int):
            new.set(value >> other)
        elif isinstance(other, (Register, Immediate)):
            new.set(value >> other.value)
        else:
            raise TypeError("unsupported operand type for Register >>: {}".format(other.__class__))
        return new

    def __invert__(self):
        new = Register(self.bits)
        new.set(self.value ^ self.mask)
        return new


class RegisterFile(object):
    def __init__(self, num: int, bits: int, immutable: list = {}, prefix: str = "x"):
        self.num = num
        self.bits = bits
        self.regs = []
        self.regs_updates = []
        for i in range(num):
            self.regs.append(Register(bits))

        for r in immutable.items():
            self.regs[r[0]].set(r[1])
            self.regs[r[0]].set_immutable(True)

        self.prefix = prefix

    def randomize(self):
        for i in range(self.num):
            self.regs[i].randomize()

    def __setitem__(self, key, value):
        if not self.regs[key].immutable:
            reg = Register(self.bits)
            reg.set(value)
            self.regs_updates.append(TraceIntegerRegister(key, reg, prefix=self.prefix, width=self.bits))

    def __getitem__(self, item):
        return self.regs[item]

    def commit(self):
        for t in self.regs_updates:
            self.regs[t.id].set(t.value)
        self.regs_updates.clear()

    def changes(self) -> list:
        return self.regs_updates.copy()

    def __str__(self):
        return "{}".format([str(r) for r in self.regs])

class SingleRegister(object):
    def __init__(self, bits: int, id: str):
        self.bits = bits
        self.reg = Register(bits)
        self.reg_update = None
        self.id = id

    def randomize(self):
        self.reg.randomize()

    def update(self, value):
        reg = Register(self.bits)
        reg.set(value)
        self.reg_update = TraceRegister(self.id, reg)

    def get(self):
        return self.reg

    def commit(self):
        if self.reg_update:
            self.reg.set(self.reg_update.value)
        self.reg_update = None

    def changes(self) -> list:
        return [self.reg_update] if self.reg_update else []

    def __int__(self):
        return self.reg.value

    def __iadd__(self, other):
        self.update(self.reg.value + other)
        return self.reg

class BitflagRegister():
    def __init__(self, flags: list, *, prefix=""):
        assert isinstance(flags, list) and len(flags) > 0
        super().__setattr__("flags", flags)
        self.prefix = prefix
        for flag in flags:
            super().__setattr__(flag, 0)
            super().__setattr__(flag+"_update", None)
    def __setattr__(self, name, value):
        if name in self.flags:
            super().__setattr__(name+"_update", value)
        else:
            super().__setattr__(name, value)

    def set(self, value):
        for flag in value:
            assert flag in self.flags
            setattr(self, flag, value[flag])

    def get(self, flag):
        assert flag in self.flags
        return getattr(self, flag)

    def changes(self):
        changes = []
        for flag in self.flags:
            upd = getattr(self, flag+"_update")
            if upd is not None:
                changes.append(TraceRegister(self.prefix+flag, upd))
        return changes

    def commit(self):
        for flag in self.flags:
            upd = getattr(self, flag+"_update")
            if upd is not None:
                super().__setattr__(flag, upd)
                super().__setattr__(flag+"_update", None)


class Trace(object):
    pass


class TracePC(Trace):
    def __init__(self, pc):
        self.pc = pc

    def __str__(self):
        return "pc = {}".format(self.pc)


class TraceRegister(Trace):
    def __init__(self, id, value):
        self.id = id
        self.value = value

    def __str__(self):
        return "{} = {}".format(self.id, self.value)


class TraceIntegerRegister(TraceRegister):
    def __init__(self, id, value, *, prefix: str="x", width=32):
        super().__init__(id, value)
        self.prefix = prefix
    def __str__(self):
        return "{}{} = {}".format(self.prefix, self.id, str(self.value))


class TraceMemory(Trace):
    GRANULARITY = Enum('granularity', ['BYTE', 'HALFWORD', 'WORD'])

    def __init__(self, gran: GRANULARITY, addr: int, data: int):
        self.gran = gran
        self.addr = addr
        self.data = data

    def __str__(self):
        if self.gran == TraceMemory.GRANULARITY.BYTE:
            data = "{:02x}".format(self.data & 0xFF)
        elif self.gran == TraceMemory.GRANULARITY.HALFWORD:
            data = "{:04x}".format(self.data & 0xFFFF)
        else:
            data = "{:08x}".format(self.data)
        return "mem[{}] = {}".format(self.addr, data)


RVFISignals = namedtuple("RVFISignals", ["valid", "order", "insn",
                                         "rs1_addr", "rs1_rdata", "rs2_addr", "rs2_rdata", "rd_addr", "rd_wdata",
                                         "pc_rdata", "pc_wdata"])
RVFISignals.__new__.__defaults__ = (None,) * len(RVFISignals._fields)
